@@ -35,6 +35,10 @@ class DummyController
   def self.controller_path
     ''
   end
+
+  def controller_path
+    ''
+  end
 end
 
 class TemplateTest < Test::Unit::TestCase
@@ -91,7 +95,8 @@ class TemplateTest < Test::Unit::TestCase
   end
 
   def assert_renders_correctly(name, &render_method)
-    if ActionPack::VERSION::MAJOR < 2 || ActionPack::VERSION::MINOR < 2
+    if ActionPack::VERSION::MAJOR < 2 ||
+        (ActionPack::VERSION::MAJOR == 2 && ActionPack::VERSION::MINOR < 2)
       render_method ||= proc { |name| @base.render(name) }
     else
       render_method ||= proc { |name| @base.render(:file => name) }
@@ -135,8 +140,16 @@ class TemplateTest < Test::Unit::TestCase
   end
 
   def test_action_view_templates_render_correctly
-    @base.instance_variable_set("@content_for_layout", 'Lorem ipsum dolor sit amet')
-    assert_renders_correctly 'content_for_layout'
+    proc = lambda do
+      @base.content_for(:layout) {'Lorem ipsum dolor sit amet'}
+      assert_renders_correctly 'content_for_layout'
+    end
+
+    if @base.respond_to?(:with_output_buffer)
+      @base.with_output_buffer("", &proc)
+    else
+      proc.call
+    end
   end
 
   def test_instance_variables_should_work_inside_templates
